/* Copyright 2015 Jetperch LLC.  All rights reserved. */

/**
 * @file
 *
 * @brief Holding place for commonly used C macros.
 */

#ifndef EMBC_CDEF_H_
#define EMBC_CDEF_H_

/**
 * @ingroup embc
 * @defgroup cdef Common C utilities
 *
 * @brief Common C macros that are not built into standard C.
 *
 * @{
 */

/**
 * @brief Perform a compile-time check
 *
 * @param COND The condition which should normally be true.
 * @param MSG The error message which must be a valid C identifier  This
 *      message will be cryptically displayed by the compiler on error.
 */
#define STATIC_ASSERT(COND, MSG) typedef char static_assertion_##MSG[(COND)?1:-1]

/**
 * @brief Compute the number of elements in an array.
 *
 * @param x The array (NOT a pointer).
 * @return The number of elements in the array.
 */
#define ARRAY_SIZE(x) ( sizeof(x) / sizeof((x)[0]) )

/**
 * @brief Evaluate and return on error.
 *
 * @param x The value to evaluate.  Any non-zero value is presumed to be an error.
 *
 * On error, cause the calling function to return with the error code.
 */
#define RETURN_ON_ERROR(x) {int rc__ = (x); if (rc__) { return rc__;} }

/**
 * @brief Evaluate and return when the expression is non-zero.
 *
 * @param x The value to evaluate.  Any non-zero value is presumed to be an error.
 * @see RETURN_ON_ERROR
 *
 * On error, cause the calling function to return with the error code.
 */
#define RWE(x) RETURN_ON_ERROR(x)

/**
 * @brief Round a value up to the nearest multiple.
 *
 * @param x The integer value to round up.
 * @param m The multiple for rounding.
 * @return The value of x rounded up to m.
 *
 * Some examples:
 *
 *     ROUND_UP_TO_MULTIPLE(1, 128) => 128
 *     ROUND_UP_TO_MULTIPLE(128, 128) => 128
 *     ROUND_UP_TO_MULTIPLE(129, 128) => 256
 */
#define ROUND_UP_TO_MULTIPLE(x, m) (( ((x) + m - 1) / m) * m)


#ifdef __cplusplus

/**
 * @brief Perform compile-time sizeof print (causes compilation to fail!)
 *
 * @param t The structure, class or type for the print.
 */
#define STATIC_PRINT_SIZEOF(t) \
    template<int s> struct sizeof_ ## t ## _is ; \
    struct sizeof_ ## t ## _is  <sizeof(t)> t ## _sz

/**
 * @brief Perform compile-time print of an integer (causes compilation to fail!)
 *
 * @param t The integer for the print.
 */
#define STATIC_PRINT_INT(i, name) \
    template<int s> struct int_ ## name ## _is; \
    struct int_ ## name ## _is <i> name ## _sz;

#endif

/** @} */

#endif /* EMBC_CDEF_H_ */
